(* $Id: hydro_locator.ml 15383 2007-11-27 18:56:43Z gerd $ *)

open Hydro_types
open Hydro_builtin

let scan_indirect_proxy s =
  try
    let k = String.index s '@' in
    `Adapter(String.sub s 0 k, String.sub s (k+1) (String.length s - k - 1))
  with
    | Not_found ->
	`WellKnownObject s

let proxy_resolver ?domain_resolver cp pr_loc =
  let base_resolver =
    Hydro_proxy.proxy_resolver ?domain_resolver cp in
  ( object(self)
      method resolve eps emit env =
	match eps#parameters with
	  | `Endpoints _ | `Connectors _ ->
	      (* Use base_resolver: *)
	      base_resolver # resolve eps emit env
	  | `Indirect s ->
	      let env' =
		Hydro_builtin_util.derive_proxy_env base_resolver env in
	      let po =
		pc_Ice_Locator env' pr_loc in
	      ( match scan_indirect_proxy s with
		  | `WellKnownObject id_str ->
		      (* CHECK: Is this correct? *)
		      let id = Hydro_builtin_util.identity_of_string id_str in
		      (po # findObjectById
			 { i_name = id#name;
			   i_category = id#category
			 }
		      ) # acall (self#process_reply None emit env)
		  | `Adapter(id_str, adapter_str) ->
		      let id = Hydro_builtin_util.identity_of_string id_str in
		      (po # findAdapterById adapter_str)
			# acall (self#process_reply (Some id) emit env)
	      )
      method private process_reply inject_id_opt emit env r =
	try
	  let found_pr =
	    match r#result with
	      | Some pr -> pr
	      | None -> (* TODO *)
		  failwith "Inactive object/adapter" in
	  let eps' =
	    (Hydro_lm.Unsafe.unwrap_proxy found_pr :> 
	       Hydro_proxy.extended_proxy_addr ) in
	  let eps'' =
	    match inject_id_opt with
	      | None -> eps' 
	      | Some id ->
		  ( object
		      method id =         id
		      method facet =      eps' # facet
		      method mode =       eps' # mode
		      method secure =     eps' # secure
		      method parameters = eps' # parameters
		    end : Hydro_proxy.extended_proxy_addr
		  ) in
	  base_resolver # resolve eps'' emit env
	with
	  | Hydro_builtin.User_exception ue 
	      when ue#hydro_ops#is_Ice_AdapterNotFoundException ->
	        emit (lazy [])
	  | Hydro_builtin.User_exception ue 
	      when ue#hydro_ops#is_Ice_ObjectNotFoundException ->
	        emit (lazy [])
	  | err -> 
	        emit (lazy (raise err))
	      
    end : Hydro_proxy.proxy_resolver_t
  )


let get_Ice_Locator_of_port host port =
  let addr =
    ( object
	method id =
	  ( object 
	      method name = "Locator"
	      method category = "IceGrid" 
	    end 
	  )
	method facet = None
	method mode = `Twoway
	method secure = false
	method parameters =
	  `Endpoints [| `TCP ( object
				 method host = host
				 method port = port
				 method timeout = 5l   (* TODO *)
				 method compress = false
			       end : tcp_endpoint
			     )
		     |]
      end : proxy_addr
    ) in
  unchecked_pr_Ice_Locator (Hydro_lm.pr_of_address addr)
