(* $Id: hydro_marshal.ml 16195 2008-01-21 22:15:59Z gerd $ *)

open Hydro_types
open Hydro_util
open Hydro_prelim

let add_char b c =
  (* TODO: forgot to export Netbuffer.add_char *)
  Netbuffer.insert_char b (Netbuffer.length b) c


let patch_buffer b pos s =
  (* TODO: functions to overwrite are missing in Netbuffer *)
  let bs = Netbuffer.unsafe_buffer b in
  String.blit s 0 bs pos (String.length s)


let print_short b x =
  let s = String.create 2 in
  let n1 = (x asr 8) land 0xff in
  String.unsafe_set s 1 (Char.unsafe_chr n1);
  let n0 = x land 0xff in
  String.unsafe_set s 0 (Char.unsafe_chr n0);
  Netbuffer.add_string b s


let encoded_int x =
  let s = String.create 4 in
  let n3 = (x asr 24) land 0xff in
  String.unsafe_set s 3 (Char.unsafe_chr n3);
  let n2 = (x lsr 16) land 0xff in
  String.unsafe_set s 2 (Char.unsafe_chr n2);
  let n1 = (x lsr 8) land 0xff in
  String.unsafe_set s 1 (Char.unsafe_chr n1);
  let n0 = x land 0xff in
  String.unsafe_set s 0 (Char.unsafe_chr n0);
  s


let print_int b x =
  let s = String.create 4 in
  let n3 = (x asr 24) land 0xff in
  String.unsafe_set s 3 (Char.unsafe_chr n3);
  let n2 = (x lsr 16) land 0xff in
  String.unsafe_set s 2 (Char.unsafe_chr n2);
  let n1 = (x lsr 8) land 0xff in
  String.unsafe_set s 1 (Char.unsafe_chr n1);
  let n0 = x land 0xff in
  String.unsafe_set s 0 (Char.unsafe_chr n0);
  Netbuffer.add_string b s


let print_int32 b x =
  let s = String.create 4 in
  let n3 = Int32.to_int (Int32.shift_right_logical x 24) land 0xff in
  String.unsafe_set s 3 (Char.unsafe_chr n3);
  let n2 = Int32.to_int (Int32.shift_right_logical x 16) land 0xff in
  String.unsafe_set s 2 (Char.unsafe_chr n2);
  let n1 = Int32.to_int (Int32.shift_right_logical x 8) land 0xff in
  String.unsafe_set s 1 (Char.unsafe_chr n1);
  let n0 = Int32.to_int (Int32.logand x 0xffl) in
  String.unsafe_set s 0 (Char.unsafe_chr n0);
  Netbuffer.add_string b s


let print_int64 b x =
  let s = String.create 8 in
  let n7 = Int64.to_int (Int64.logand (Int64.shift_right_logical x 56)
                           0xffL) in
  String.unsafe_set s 7 (Char.unsafe_chr n7);
  let n6 = Int64.to_int (Int64.logand (Int64.shift_right_logical x 48)
                           0xffL) in
  String.unsafe_set s 6 (Char.unsafe_chr n6);
  let n5 = Int64.to_int (Int64.logand (Int64.shift_right_logical x 40)
                           0xffL) in
  String.unsafe_set s 5 (Char.unsafe_chr n5);
  let n4 = Int64.to_int (Int64.logand (Int64.shift_right_logical x 32)
                           0xffL) in
  String.unsafe_set s 4 (Char.unsafe_chr n4);
  let n3 = Int64.to_int (Int64.logand (Int64.shift_right_logical x 24)
                           0xffL) in
  String.unsafe_set s 3 (Char.unsafe_chr n3);
  let n2 = Int64.to_int (Int64.logand (Int64.shift_right_logical x 16)
                           0xffL) in
  String.unsafe_set s 2 (Char.unsafe_chr n2);
  let n1 = Int64.to_int (Int64.logand (Int64.shift_right_logical x 8)
                           0xffL) in
  String.unsafe_set s 1 (Char.unsafe_chr n1);
  let n0 = Int64.to_int (Int64.logand x 0xffL) in
  String.unsafe_set s 0 (Char.unsafe_chr n0);
  Netbuffer.add_string b s


let print_float b x =
  print_int32 b (Int32.bits_of_float x)


let print_double b x =
  print_int64 b (Int64.bits_of_float x)


let print_size b n =
  if n < 255 then
    add_char b (Char.unsafe_chr n)
  else (
    add_char b '\255';
    print_int b n
  )


let encapsulate f b e =
  let p = Netbuffer.length b in
  Netbuffer.add_string b "\000\000\000\000\001";
  Netbuffer.add_char b (Char.chr e);
  f b e;
  let p' = Netbuffer.length b in
  let delta = encoded_int (p' - p) in
  patch_buffer b p delta

let encapsulate2 b1 eb2 =
  (* Add the encapsulation header for existing eb2 to the end of b1 *)
  let l = eb2.encap_len + 6 in
  print_int b1 l;
  Netbuffer.add_string b1 "\001";
  Netbuffer.add_char b1 (Char.chr eb2.encap_enc_minor);
  { encap_buf = b1;
    encap_pos = 0;
    encap_len = Netbuffer.length b1;
    encap_enc_minor = eb2.encap_enc_minor
  }

type marshal_env =
    { cmap : (int, string * object_value * int) Hashtbl.t;
        (* Maps id to (class_name, class_value, inst_id) where
           - class_name is the absolute type ID of the class,
           - class_value is the object so that [Oo.id class_value = id]
           - inst_id is the instance ID used in the serialized string
         *)
      cq : int Queue.t;
        (* The queue of class value IDs that needs to be appended *)
      next_inst_id : int ref;
        (* The next instance ID to allocate *)
      tmap : int CiHashtbl.t;
        (* Maps absolute class type IDs to numerical values *)
      next_type_id : int ref;
        (* The next type ID to allocate *)
      em : int;
        (* Minor encoding version *)
    }


let rec marshal_value env (sys:system) ht v b =
  match (ht, v) with
    | (_, `DirectWriter f) ->
	f b
    | (`Void, `Nothing) ->
	()
    | (`Bool, `Bool flag) ->
	add_char b (if flag then '\001' else '\000')
    | (`Byte, `Byte n) ->
	add_char b (Char.unsafe_chr (n land 0xff))
    | (`Short, `Short n) ->
	print_short b n
    | ((`Int|`Int32), `Int32 n) ->
	print_int32 b n
    | ((`Int|`Int32), `Int n) ->
	(* CHECK: Marshalling error if n out of range on 64 bit platforms? *)
	print_int b n
    | (`Long, `Long n) ->
	print_int64 b n
    | (`Float, `Float x) ->
	print_float b x
    | (`Double, `Double x) ->
	print_double b x
    | (`String, `String s) ->
	print_size b (String.length s);
	Netbuffer.add_string b s
    | ((`Sequence `Byte|`Byteseq), `Byteseq s) ->
	print_size b (String.length s);
	Netbuffer.add_string b s
    | (`Sequence ht', `Sequence seq) ->
	print_size b (Array.length seq);
	Array.iter
	  (fun v' -> marshal_value env sys ht' v' b)
	  seq
    | (`Dictionary(ht1, ht2), `Dictionary seq) ->
	print_size b (Array.length seq);
	Array.iter
	  (fun (v1,v2) -> 
	     marshal_value env sys ht1 v1 b;
	     marshal_value env sys ht2 v2 b
	  )
	  seq
    | (`Enum e, `Enum k) ->
	let n = Array.length e in
	if k < 0 || k >= n then
	  raise(Marshal_error "Enum out of range");
	if n <= 127 then
	  add_char b (Char.unsafe_chr (k land 0xff))
	else
	  if n <= 32767 then
	    print_short b k
	  else
	    print_int b k
    | (`Struct st, `Struct s) ->
	if Array.length st <> Array.length s then
	  raise(Marshal_error "Struct does not match definition");
	Array.iteri
	  (fun k v' ->
	     let (_, ht') = st.(k) in
	     marshal_value env sys ht' v' b
	  )
	  s
    | (`Class cname, `Class cvalref) ->
	( match !cvalref with
	    | `Placeholder _ ->
		raise(Marshal_error "Cannot marshal class placeholder")
	    | `Value cval ->
		let id = Oo.id cval in
		try
		  let (_, _, inst_id) = Hashtbl.find env.cmap id in
		  (* The class value is already known, and is either already
                     marshalled, or scheduled for marshalling
                   *)
		  print_int b (-inst_id)
		with
		  | Not_found ->
		      (* The class value was not seen before *)
		      let inst_id = !(env.next_inst_id) in
		      incr env.next_inst_id;
		      Hashtbl.add env.cmap id (cname, cval, inst_id);
		      Queue.add id env.cq;
		      print_int b (-inst_id)
	)

    | (`Class _, `Null) ->
	print_int b 0

    | (`Proxy name, `Proxy p) ->
	if not (CiHashtbl.mem sys#interfaces name) then
	  raise(Marshal_error("Proxy not found: " ^ name));
	if name = "" then
	  raise(Marshal_error("Cannot marshal proxies with empty name"));
	marshal_value env sys t_proxy
	  (`Struct
	     [| `String p#id#name;
		`String p#id#category;
		`Sequence (match p#facet with
			     | None -> [| |]
			     | Some fc -> [| `String fc |]
			  );
		`Byte (match p#mode with
			 | `Twoway -> 0
			 | `Oneway -> 1
			 | `Batch_oneway -> 2
			 | `Datagram -> 3
			 | `Batch_datagram -> 4
		      );
		`Bool p#secure;
	     |]
	  ) b;
	( match p#parameters with
	    | `Endpoints eps ->
		if Array.length eps = 0 then
		  raise(Marshal_error "A proxy with endpoints must have at least one endpoint");
		print_size b (Array.length eps);
		Array.iter
		  (fun ep ->
		     (* Note: The spec says that the endpoints are in their
                        own encapsulations. As the endpoint encoding does
                        not use class values, we can reuse [env].
                      *)
		     match ep with
		       | `TCP ep ->
			   print_short b 1;
			   encapsulate
			     (fun b _ ->
				marshal_value env sys t_tcp_endpoint
				  (`Struct [| `String ep#host;
					      `Int ep#port;
					      `Int32 ep#timeout;
					      `Bool ep#compress
					   |]) b
			     )
			     b
			     env.em
		       | `UDP ep ->
			   print_short b 2;
			   encapsulate
			     (fun b _ ->
				marshal_value env sys t_udp_endpoint
				  (`Struct [| `String ep#host;
					      `Int ep#port;
					      `Int ep#proto_major;
					      `Int ep#proto_minor;
					      `Int ep#enc_major;
					      `Int ep#enc_minor;
					      `Bool ep#compress
					   |]) b
			     )
			     b
			     env.em
		       | `SSL ep ->
			   print_short b 3;
			   encapsulate
			     (fun b _ ->
				marshal_value env sys t_tcp_endpoint
				  (`Struct [| `String ep#host;
					      `Int ep#port;
					      `Int32 ep#timeout;
					      `Bool ep#compress
					   |]) b
			     )
			     b
			     env.em
		       | `Unknown(n,s) ->
			   print_short b n;
			   encapsulate 
			     (fun b _ -> Netbuffer.add_string b s) b env.em
		  )
		  eps;
	    | `Indirect s ->
		print_size b 0;
		print_size b (String.length s);
		Netbuffer.add_string b s
	)

    | (`Proxy name, `Null) ->
	print_size b 0;  (* empty name *)
	print_size b 0   (* empty category *)

    | _ ->
	raise(Marshal_error "Value does not match type")


let marshal_type_id env type_id b =
  try
    let num_id = CiHashtbl.find env.tmap type_id in
    add_char b '\001';
    print_size b num_id
  with Not_found ->
    let num_id = !(env.next_type_id) in
    incr env.next_type_id;
    CiHashtbl.add env.tmap type_id num_id;
    add_char b '\000';
    print_size b (String.length type_id);
    Netbuffer.add_string b type_id


let marshal_class_val env sys cname cval inst_id b =
  let cdef =
    try CiHashtbl.find sys#classes cname
    with Not_found -> raise(Marshal_error("Class not found: " ^ cname)) in
  print_int b inst_id;
  
  (* Iterate over the slices: *)
  let slices = ref (List.rev (cval#hydro_slices)) in
  let next_class = ref (Some cdef) in

  while !next_class <> None do
    match !next_class with
      | Some next_cdef ->
	  ( match !slices with
	      | (`Decoded(type_id, sval)) :: slices' ->
		   if type_id <> next_cdef#name then
		     raise(Marshal_error("Class definition mismatch"));
   		   marshal_type_id env type_id b;
		   let p = Netbuffer.length b in
		   Netbuffer.add_string b "\000\000\000\000";
		   marshal_value 
		     env sys (`Struct next_cdef#elements) (`Struct sval) b;
		   let p' = Netbuffer.length b in
		   let delta = encoded_int (p' - p) in
		   patch_buffer b p delta;

		   slices := slices'

	      | (`Opaque(type_id, data)) :: slices' ->
		   if type_id <> next_cdef#name then
		     raise(Marshal_error("Class definition mismatch"));
   		   marshal_type_id env type_id b;
		   let p = Netbuffer.length b in
		   Netbuffer.add_string b "\000\000\000\000";
		   Netbuffer.add_string b data;
		   let p' = Netbuffer.length b in
		   let delta = encoded_int (p' - p) in
		   patch_buffer b p delta;

		   slices := slices'

	      | [] ->
		  raise(Marshal_error("Missing slice to marshal class value"))
	  );
	  next_class := next_cdef # super
      | _ -> assert false
  done


let marshal_class_appendix env sys b =
  let cq' = Queue.create() in
  while not (Queue.is_empty env.cq) do
    let n = Queue.length env.cq in
    print_size b n;
    
    Queue.transfer env.cq cq';
    while not (Queue.is_empty cq') do
      let id = Queue.pop cq' in
      let (cname, cval, inst_id) = 
	try Hashtbl.find env.cmap id with Not_found -> assert false in
      marshal_class_val env sys cname cval inst_id b;
    done
  done;
  print_size b 0


let marshal sys ht class_flag v b em =
  if em < 0 then
    raise(Limitation `UnsupportedEncodingVersion);
  let em = 0 in  (* We can only do version 0! *)
  let cq = Queue.create() in
  let env =
    { cmap = Hashtbl.create 10;
      cq = cq;
      next_inst_id = ref 1;
      tmap = CiHashtbl.create 10;
      next_type_id = ref 1;
      em = em
    } in

  marshal_value env sys ht v b;

  if class_flag then
    marshal_class_appendix env sys b
  else
    if not (Queue.is_empty cq) then
      raise(Marshal_error("Classes found although the class flag is false"))


let marshal_exn sys hx (sv:sliced_value) b em =
  if em < 0 then
    raise(Limitation `UnsupportedEncodingVersion);
  let em = 0 in  (* We can only do version 0! *)
  let p_exn = Netbuffer.length b in
  print_size b 0;

  let cq = Queue.create() in
  let env =
    { cmap = Hashtbl.create 10;
      cq = cq;
      next_inst_id = ref 1;
      tmap = CiHashtbl.create 10;
      next_type_id = ref 1;
      em = em
    } in

  (* Iterate over the slices: *)
  let slices = List.rev sv#hydro_slices in
  let prev_eff_exn = ref None in
  let eff_type_found = ref false in
  let static_type_found = ref false in

  List.iter
    (fun slice ->
       match slice with
	 | `Decoded(type_id, sval) ->
	     (* Check value: *)
	     let eff_x =
	       try CiHashtbl.find sys#exceptions type_id
	       with Not_found ->
		 raise(Marshal_error("Exception ID not found: " ^ type_id)) in
	     if eff_x#name = hx#name then
	       static_type_found := true;
	     if not !eff_type_found then (
	       if type_id <> sv#hydro_effective_id then
		 raise(Marshal_error("Bad effective type of exception"));
	       eff_type_found := true
	     );
	     ( match !prev_eff_exn with
		 | None -> ()
		 | Some p_eff_x ->
		     let ok =
		       match p_eff_x#super with
			 | None -> false
			 | Some px -> px#name = type_id in
		     if not ok then
		       raise(Marshal_error("Exception definition mismatch"))
	     );
	     prev_eff_exn := Some eff_x;
	     (* Marshal value: *)
	     print_size b (String.length type_id);
	     Netbuffer.add_string b type_id;
	     let p = Netbuffer.length b in
	     Netbuffer.add_string b "\000\000\000\000";
	     marshal_value 
	       env sys (`Struct eff_x#elements) (`Struct sval) b;
	     let p' = Netbuffer.length b in
	     let delta = encoded_int (p' - p) in
	     patch_buffer b p delta;

	 | `Opaque(type_id, data) ->
	     (* Check value: *)
	     if !eff_type_found then
	       raise(Marshal_error("Exception contains opaque data at wrong place"));
	     (* Marshal value: *)
	     print_size b (String.length type_id);
	     Netbuffer.add_string b type_id;
	     let p = Netbuffer.length b in
	     Netbuffer.add_string b "\000\000\000\000";
	     Netbuffer.add_string b data;
	     let p' = Netbuffer.length b in
	     let delta = encoded_int (p' - p) in
	     patch_buffer b p delta;
    )
    slices;

  (* Further checks: *)
  if not !eff_type_found then
    raise(Marshal_error "Bad effective type of exception");
  if not !static_type_found then
    raise(Marshal_error "Effective type of exception does not match static type");
  ( match !prev_eff_exn with
      | None -> assert false
      | Some last_exn ->
	  if last_exn#super <> None then
	    raise(Marshal_error("Exception definition mismatch"))
  );

  if not (Queue.is_empty cq) then (
    (* If no class values have been found, we don't have to marshal them *)
    (* CHECK whether this is ok! *)
    patch_buffer b p_exn "\001";
    marshal_class_appendix env sys b
  )


let make_header zstatus mt len pm em : msg_header =
  ( object
      method proto_major = 1
      method proto_minor = pm
      method enc_major = 1
      method enc_minor = em
      method msg_type = mt
      method compression = zstatus
      method body_size = len
    end
  )


let marshal_msg sys zstatus (msg:msg) pm em =
  if pm < 0 then
    raise(Limitation `UnsupportedProtocolVersion);
  if em < 0 then
    raise(Limitation `UnsupportedEncodingVersion);
  let pm = 0 in  (* We can only do version 0! *)
  let em = 0 in  (* We can only do version 0! *)
  match msg with
    | `Request req ->
	let v = 
	  `Struct [| `Int32 req#request_id;
		     `String req#id#name;
		     `String req#id#category;
		     `Sequence ( match req#facet with
				   | None -> [| |]
				   | Some fc -> [| `String fc |]
			       );
		     `String req#operation;
		     `Byte (match req#mode with
			      | `Normal -> 0
			      | `Nonmutating -> 1
			      | `Idempotent -> 2
			   );
		     `Dictionary (Array.map 
				    (fun (k,v) -> (`String k, `String v))
				    req#context
				 );
		  |] in
	let b1 = Netbuffer.create 200 in
	marshal sys t_msg_request false v b1 em;
	let eb2 = req#params in
	let eb1 = encapsulate2 b1 eb2 in
	let total_len = eb1.encap_len + eb2.encap_len in
	let hdr = make_header zstatus `Request total_len pm em in
	(hdr, [ eb1; eb2 ] )

    | `Batch_request reqs ->
	let b0 = Netbuffer.create 10 in
	print_int b0 (List.length reqs);
	let bufs =
	  List.flatten
	    (List.map
	       (fun req ->
		  let dict =
		    Array.map 
		      (fun (k,v) -> (`String k, `String v))
		      req#context in
		  let v = 
		    `Struct [| `String req#id#name;
			       `String req#id#category;
			       `Sequence ( match req#facet with
					     | None -> [| |]
					     | Some fc -> [| `String fc |]
					 );
			       `String req#operation;
			       `Byte (match req#mode with
					| `Normal -> 0
					| `Nonmutating -> 1
					| `Idempotent -> 2
				     );
			       `Dictionary dict;
			    |] in
		  let b1 = Netbuffer.create 200 in
		  marshal sys t_msg_batch_request false v b1 em;
		  let eb2 = req#params in
		  let eb1 = encapsulate2 b1 eb2 in
		  [ eb1; eb2 ]
	       )
	       reqs
	    ) in
	let eb0 = { encap_buf = b0; 
		    encap_pos = 0; 
		    encap_len = Netbuffer.length b0;
		    encap_enc_minor = em
		  } in
	let all_bufs = eb0 :: bufs in
	let total_len =
	  List.fold_left
	    (fun acc eb -> acc + eb.encap_len)
	    0
	    all_bufs in
	let hdr = make_header zstatus `Batch_request total_len pm em in
	(hdr, all_bufs)

    | `Reply rpy ->
	let v = 
	  `Struct [| `Int32 rpy#request_id;
		     `Byte (match rpy#result with
			      | `Success _ -> 0
			      | `User_exception _ -> 1
			      | `Object_does_not_exist _ -> 2
			      | `Facet_does_not_exist _ -> 3
			      | `Operation_does_not_exist _ -> 4
			      | `Unknown_local_exception _ -> 5
			      | `Unknown_user_exception  _ -> 6
			      | `Unknown_exception _ -> 7
			   );
		  |] in
	let b1 = Netbuffer.create 200 in
	marshal sys t_msg_reply false v b1 em;
	let eb2_opt =
	  match rpy#result with
	    | `Success eb2 
	    | `User_exception eb2 ->
		Some eb2
 	  (* The Ice manual says there is always an encapsulation.
             However, this seems not to be true for reply types
             >= 2
           *)
	    | `Object_does_not_exist(id, facet, op)
	    | `Facet_does_not_exist(id, facet, op)
	    | `Operation_does_not_exist(id, facet, op) ->
		let e = `Struct [| `String id#name;
				   `String id#category;
				   `Sequence ( match facet with
						 | None -> [| |]
						 | Some fc -> [| `String fc |]
					     );
				   `String op
				|] in
		marshal sys t_msg_reply_error false e b1 em;
		None
	    | `Unknown_local_exception s 
	    | `Unknown_user_exception s
	    | `Unknown_exception s ->
		marshal sys `String false (`String s) b1 em;
		None
	in
	( match eb2_opt with
	    | Some eb2 ->
		let eb1 = encapsulate2 b1 eb2 in
		let total_len = eb1.encap_len + eb2.encap_len in
		let hdr = make_header zstatus `Reply total_len pm em in
		(hdr, [ eb1; eb2 ])
	    | None ->
		let eb1 = 
		  { encap_buf = b1;
		    encap_pos = 0;
		    encap_len = Netbuffer.length b1;
		    encap_enc_minor = em
		  } in
		let total_len = eb1.encap_len in
		let hdr = make_header zstatus `Reply total_len pm em in
		(hdr, [ eb1 ])
	)

    | `Validate_connection ->
	let hdr = make_header `Uncompressed `Validate_connection 0 pm em in
	(hdr, [] )

    | `Close_connection ->
	let hdr = make_header `Uncompressed `Close_connection 0 pm em in
	(hdr, [] )


let max_proto_minor = 0
let max_enc_minor = 0
