(* $Id: hydro_netplex.ml 15689 2007-12-13 22:28:57Z gerd $ *)

let hydro_factory 
      ~configure
      ?(hooks = fun _ -> new Netplex_kit.empty_processor_hooks())
      ?(supported_ptypes = [ `Multi_processing ])
      ~name
      ~sys
      ~setup
      () =

  ( object(self)
      method name = name
      method create_processor ctrl_cfg cf addr =
	let custom_cfg = configure cf addr in

	let srv_list = ref [] in

	( object(self)
	    inherit Netplex_kit.processor_base (hooks custom_cfg) as super

	    method shutdown () =
	      List.iter
		(fun srv ->
		   Hydro_endpoint.Server.shutdown srv)
		!srv_list;
	      srv_list := [];
	      super # shutdown()
		

	    method process ~when_done container fd proto =
	      let esys = container # event_system in
	      let descr = 
		Hydro_connector.descriptor fd false `Stream in
	      let params = 
		Hydro_params.server_params() (* TODO *) in

	      let onabort srv =
		(* FIXME: using the onabort hook might be wrong if we allow
                   SSL, because it is called too early - when the logical
                   server circuit is aborted, and not when the last bits of
                   the connection are torn down. For TCP this is the same,
                   but for SSL the connection shutdown takes a bit longer
                   than the server lives
                 *)
		srv_list :=
		  List.filter
		    (fun srv' -> srv' != srv)
		    !srv_list;
		let g = Unixqueue.new_group esys in
		Unixqueue.once esys g 0.0 when_done
	      in
	      let srv =
		Hydro_endpoint.Server.create
		  ~onabort
		  sys
		  `TCP
		  descr
		  params
		  esys in
	      srv_list := srv :: !srv_list;
	      setup container srv custom_cfg

	    method supported_ptypes = 
	      supported_ptypes

	  end
	)
    end
  )
;;
